@@ -6,6 +6,8 @@
 #' @inheritParams trainSingleR
 #' @param de.n An integer scalar specifying the number of DE genes to use.
 #' Defaults to \code{500 * (2/3) ^ log2(N)} where \code{N} is the number of unique labels.
+#' @param num.threads Integer scalar specifying the number of threads to use.
+#' @param BPPARAM A \linkS4class{BiocParallelParam} object specifying how parallelization should be performed.
 #'
 #' @return
 #' A list of lists of character vectors, 
@@ -48,8 +50,9 @@
 #'
 #' @export
 #' @importFrom S4Vectors selfmatch DataFrame
-#' @importFrom utils head
-getClassicMarkers <- function(ref, labels, assay.type="logcounts", check.missing=TRUE, de.n=NULL, BPPARAM=SerialParam()) { 
+#' @importFrom BiocParallel SerialParam bpnworkers
+#' @importFrom utils relist
+getClassicMarkers <- function(ref, labels, assay.type="logcounts", check.missing=TRUE, de.n=NULL, num.threads=bpnworkers(BPPARAM), BPPARAM=SerialParam()) { 
     if (!bpisup(BPPARAM) && !is(BPPARAM, "MulticoreParam")) {
         bpstart(BPPARAM)
         on.exit(bpstop(BPPARAM))
@@ -65,59 +68,65 @@ getClassicMarkers <- function(ref, labels, assay.type="logcounts", check.missing
     if (length(common)==0L && any(vapply(ref, nrow, 0L) > 0L)) {
         stop("no common row names across 'ref'")
     }
+    common <- as.character(common) # avoid problems with NULL rownames for zero-row inputs.
 
     for (i in seq_along(ref)) {
         current <- ref[[i]][common,,drop=FALSE]
         current <- .to_clean_matrix(current, assay.type, check.missing, msg="ref", BPPARAM=BPPARAM)
         ref[[i]] <- .median_by_label(current, labels[[i]], BPPARAM=BPPARAM)
     }
-
-    # Identify all label combinations within each reference.
-    collated <- list()
-    for (i in seq_along(ref)) {
-        curnames <- as.character(colnames(ref[[i]])) # coerce NULL to character(0) for zero-ncol ref.
-        pairs <- expand.grid(first=curnames, second=curnames, stringsAsFactors=FALSE)
-        pairs <- pairs[pairs$first!=pairs$second,]
-        collated[[i]] <- DataFrame(pairs, index=rep(i, nrow(pairs)))
-    }
-
-    collated <- do.call(rbind, collated)
-    chosen <- selfmatch(collated[,1:2])
-    indices <- split(collated$index, chosen)
-    choices <- collated[match(as.integer(names(indices)), chosen),]
-
-    # Setting up the output structure.
+    
     ulabels <- .get_levels(unlist(lapply(ref, colnames)))
-    output <- vector("list", length(ulabels))
-    names(output) <- ulabels
-    empty <- output
-
-    for (i in ulabels) {
-        subout <- empty
-        for (j in ulabels) {
-            subout[[j]] <- character(0)
-        }
-        output[[i]] <- subout
+    labels <- list()
+    for (i in seq_along(ref)) {
+        m <- match(colnames(ref[[i]]), ulabels)
+        labels[[i]] <- m - 1L
     }
 
     # Identify top hits based on the average (or sum, it doesn't matter)
     # of the log-fold changes between labels across references.
     if (is.null(de.n)) {
-        de.n <- round(500*(2/3)^log2(length(ulabels)))
+        de.n <- -1L
+    } else {
+        stopifnot(de.n > 0)
     }
 
-    for (i in seq_along(indices)) {
-        left <- choices$first[i]
-        right <- choices$second[i]
+    out <- find_classic_markers(nlabels=length(ulabels), 
+        ngenes=length(common), 
+        labels=labels,
+        ref=ref, 
+        de_n=de.n,
+        nthreads=num.threads
+    )
 
-        lfc <- 0
-        for (r in indices[[i]]) {
-            lfc <- lfc + (ref[[r]][,left] - ref[[r]][,right])
-        }
-
-        keep <- head(order(lfc, decreasing=TRUE), de.n)
-        output[[left]][[right]] <- as.character(names(lfc)[keep]) # to handle NULL rownames upon matrix zero-subsetting.
+    names(out) <- ulabels
+    for (i in seq_along(out)) {
+        names(out[[i]]) <- ulabels
     }
 
+    relist(common[unlist(out)], out)
+}
+
+#' @importFrom DelayedMatrixStats rowMedians
+#' @importFrom DelayedArray DelayedArray
+.median_by_label <- function(mat, labels, BPPARAM=SerialParam()) {
+    old <- getAutoBPPARAM()
+    setAutoBPPARAM(BPPARAM)
+    on.exit(setAutoBPPARAM(old))
+
+    if (!bpisup(BPPARAM) && !is(BPPARAM, "MulticoreParam")) {
+        bpstart(BPPARAM)
+        on.exit(bpstop(BPPARAM))
+    }
+
+    ulabels <- .get_levels(labels)
+    output <- matrix(0, nrow(mat), length(ulabels))
+    rownames(output) <- rownames(mat)
+    colnames(output) <- ulabels
+
+    for (u in ulabels) {
+        # Disambiguate from Biobase::rowMedians.
+        output[,u] <- DelayedMatrixStats::rowMedians(DelayedArray(mat), cols=u==labels)
+    }
     output
 }
